/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package hu.sztaki.dsd.visionair.a2a3d;

import hu.sztaki.dsd.visionair.a2a3d.agentCommunication.WebInterface;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSource;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSourceFactory;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.NoEventSourceAviableError;
import hu.sztaki.dsd.visionair.a2a3d.localStorage.DataStore;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.XMLConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>
 * A2A3D represents the main class of A2A3D Agents. It implements the
 * functionality of the configurator, the Sensor Trigger Repository, the Data
 * Collector Repository, and initialises the trigger queues of the data
 * collectors. The configuration of the sensor is stored in the A2A3DSetup.xml
 * XML file.
 *
 *
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt; (MTA
 * SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
public class A2A3D {

    /**
     * Status of the agent during it's life-cycle: uninitialised.
     * 
     * <p>This is the 1st state, the service is not available neither the 
     * event sources are running.
     */
    public static int A2A3D_STATE_UNINITIALISED = 0;

    /**
     * Status of the agent during it's life-cycle: initialising.
     * 
     * <p>This is the 2nd state, the service is not available neither the 
     * event sources are running.
     */
    public static int A2A3D_STATE_INITIALISING = 1;

    /**
     * Status of the agent during it's life-cycle: initialising the web interface.
     * 
     * <p>This is the 3rd state, the service is not available neither the 
     * event sources are running. The web interface is being initialised.
     */
    public static int A2A3D_STATE_INITINGWEBINTERFACE = 3;

    /**
     * Status of the agent during it's life-cycle: initialising the event sources.
     * 
     * <p>This is the 4th state, the service is available, the event sources
     * are being initialised one after another.
     */
    public static int A2A3D_STATE_ADDINGEVENTSOURCES = 4;

    /**
     * Status of the agent during it's life-cycle: running.
     * 
     * <p>This is the 5th state, the service is available, the event sources
     * are up and running, too.
     */
    public static int A2A3D_STATE_RUNNING = 5;

    /**
     * Status of the agent during it's life-cycle: stopping.
     * 
     * <p>This is the 6th state, the service is being shut down, the event 
     * sources are being shut down one after the another.
     */
    public static int A2A3D_STATE_STOPPING = 6;

    /**
     * Status of the agent during it's life-cycle: stopped.
     * 
     * <p>This is the 7th state, the service is shut down.
     */
    public static int A2A3D_STATE_STOPPED = 7;

    /**
     * Logger of the class.
     */
    private static Logger logger = LoggerFactory.getLogger(A2A3D.class);
    /**
     * Communication and administration layer to the agent.
     */
    public WebInterface webInterface;
    /**
     * Configuration settings. The configuration file (A2A3DSetup.xml) must be
     * in the classpath.
     */
    private XMLConfiguration xmlConfiguration;
    /**
     * HashMap of EventSources.
     */
    private HashMap<String, EventSource> eventSourceMap = new HashMap<String, EventSource>();

    /**
     * Persistent local database.
     */
    private DataStore localStore;
    /**
     * The running instance of the A2A3D.
     */
    public static A2A3D instance;
    /**
     * This thread keeps alive the sensor.
     */
    private static Thread mainThread;

    /**
     * The id of the machine. It is used to distinguish the event sources from
     * different machines. While in this version the uniqueness of this is not 
     * checked, it is considered to be unique.
     */
    public static String machineId;

    private final Object mainLock = new Object();
    private final String configFile;
    /**
     * The status of the agent. Legal values are: A2A3D_STATE_UNINITIALISED,
     * A2A3D_STATE_INITIALISING
     */
    public static int state = A2A3D_STATE_UNINITIALISED;

    /**
     * The main class of the A2A3D.
     *
     * @param args the first argument might be the path of the config file.
     * <p>If it is missing, the default A2A3DSetup.xml is used from the 
     * classpath.
     */
    public static void main(String[] args) {
        String configFile = null;
        if (!args.equals(null)) {
            configFile = args[0];
        } else {
            configFile = "A2A3DSetup.xml";
        }
        instance = new A2A3D(configFile);
        instance.init();
        instance.run();
    }

    /**
     * Creates a new instance of the A2A3D. Reads the configuration file, loads
     * the trigger event sources and the data collectors into their
     * repositories, based on the configuration fills the queue of the the
     * triggers and keeps alive the sensor application. When the application
     * exits, it will call the destroy method of the event sources and triggers.
     * @param config
     */
    public A2A3D(String config) {
        configFile = config;
    }

    /**
     * Initialises and starts the agent.
     */
    private void init() {
        try {
            state = A2A3D_STATE_INITIALISING;
            logger.info("A2A3D agent is being initialised");
            logger.info("Config file: " + configFile);

            mainThread = Thread.currentThread();

            // Load the configuration file
            xmlConfiguration = new XMLConfiguration(configFile);
            machineId = xmlConfiguration.configurationAt("Communication.ServerSocket").getString("MachineId");
            // Instantiate the DataStore
            localStore = new DataStore();
            state = A2A3D_STATE_INITINGWEBINTERFACE;
            webInterface = new WebInterface(xmlConfiguration.configurationAt("Communication.ServerSocket"));

            // Add a shutdown hook to properly close the data collectors when
            // terminating the sensor.
            Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {

                @Override
                public void run() {
                    logger.debug("SHUTDOWN HOOK RUNNING");
                    close();
                    logger.info("Shutdown hook finished.");
                }
            }));

            state = A2A3D_STATE_ADDINGEVENTSOURCES;
            // Look for event sources
            List eventSourceList = xmlConfiguration.getList("EventSources.EventSource.ClassName");
            logger.debug("Number of found eventSources: " + eventSourceList.size());

            // Instantiate all EventSource set in the configuration.
            for (int listCount = 0; listCount < eventSourceList.size(); listCount++) {
                try {
                    Class eventSourceFactoryClass = Class.forName("" + eventSourceList.get(listCount));
                    if (EventSourceFactory.class.isAssignableFrom(eventSourceFactoryClass)) {
                        String sourceId = xmlConfiguration.getString("EventSources.EventSource(" + listCount + ").Id");
                        if (!eventSourceMap.containsKey(sourceId)) {
                            // the eventsource is not yet in the cache
                            try {
                                logger.info("Initializing a new EventSource: " + sourceId);
                                EventSourceFactory esf = (EventSourceFactory) eventSourceFactoryClass.newInstance();
                                EventSource es;
                                try {
                                    es = esf.getInstance(xmlConfiguration.configurationAt("EventSources.EventSource(" + listCount + ")"));
                                    es.init();
                                    eventSourceMap.put(sourceId, es);
                                    logger.info("Initializing of eventSource: " + sourceId + " - Finished");
                                } catch (NoEventSourceAviableError ex) {
                                    logger.error(ex.toString());
                                }
                            } catch (InstantiationException ex) {
                                logger.error(ex.toString());
                            } catch (IllegalAccessException ex) {
                                logger.error(ex.toString());
                            }
                        } else {
                            logger.error("Invalid EventSoure id(" + sourceId + "). The same id is already stored in the repository.");
                            System.exit(1);
                        }
                    }
                } catch (ClassNotFoundException ex) {
                    logger.error(ex.toString());
                    logger.error("Invalid EventSource className.");
                    System.exit(1);
                }
            }
            
            // Fill in the notification queues of the EventSource-s
            for (int listCount = 0; listCount < eventSourceList.size(); listCount++) {
                String sourceId = xmlConfiguration.getString("EventSources.EventSource(" + listCount + ").Id");
                List triggerList = xmlConfiguration.getList("EventSources.EventSource(" + listCount + ").Listens.Trigger");
                logger.debug("The size of the triggerlist: " + triggerList.size());
                for (int tListCount = 0; tListCount < triggerList.size(); tListCount++) {
                    EventSource srcInMap = (EventSource) eventSourceMap.get(triggerList.get(tListCount));
                    if (srcInMap != null) {
                        srcInMap.addToQueue(sourceId, eventSourceMap.get(sourceId));
                        logger.info("The EventSource has been added to the queue of EventSource " + triggerList.get(tListCount));
                    } else {
                        logger.error("This EventSource required as trigger source for " + sourceId + " is missing: " + triggerList.get(tListCount));
                        System.exit(1);
                    }
                }
            }
            state = A2A3D_STATE_RUNNING;
            logger.debug("INIT FINISHED");
        } catch (ConfigurationException ex) {
            java.util.logging.Logger.getLogger(A2A3D.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private void run() {
        logger.info("The agent is running");
        while (true) {
            try {
                synchronized (mainLock) {
                    mainLock.wait();
                }
            } catch (InterruptedException ex) {
                logger.debug(ex.toString());
                logger.info("The agent is terminating");
                close();
                break;
            }
        }
        logger.info("The agent is stopped");

    }

    /**
     * Terminates the sensor.
     */
    public static void stop() {
        logger.debug("Stopping the agent ..");
        close();
        mainThread.interrupt();
    }

    /**
     * Returns the instance of the running A2A3D.
     *
     * @return A2A3D instance
     */
    public static A2A3D getInstance() {
        return instance;
    }

    private static void close() {
        state = A2A3D_STATE_STOPPING;
        logger.debug("Close running.");
        A2A3D actInstance = A2A3D.getInstance();
        Iterator eventSourceIterator = actInstance.eventSourceMap.keySet().iterator();
        logger.debug("Destroying the trigger event sources.");
        while (eventSourceIterator.hasNext()) {
            EventSource eventSource = (EventSource) actInstance.eventSourceMap.get(eventSourceIterator.next());
            eventSource.destroy();
        }

        actInstance.webInterface.close();
        actInstance.localStore.close();

        state = A2A3D_STATE_STOPPED;
    }

    /**
     *
     * @return
     */
    public HashMap<String, EventSource> getEventSourceMap() {
        return eventSourceMap;
    }

    /**
     *
     * @return
     */
    public DataStore getDataStore() {
        return localStore;
    }

    /*public HashMap<String, DataCollector> getDataCollectorMap() {
     return dataCollectorMap;
     }

     public EventDistributor getEventDistributor() {
     return eventDistributor;
     }*/
}
