/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package hu.sztaki.dsd.visionair.a2a3d;

import java.io.IOException;
import java.io.StringWriter;
import java.util.Date;
import javax.persistence.*;
import javax.xml.bind.annotation.XmlElement;
import org.codehaus.jackson.map.ObjectMapper;
import org.hibernate.annotations.GenericGenerator;

/**
 * <p>
 * The internal representation of the collected desktop events in the agent.
 * These events might be stored in the local SQLite database, and they can be
 * forwarded to another agent.
 *
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt;
 * (MTA SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
@Entity
@Table(name = "events")
public class Event {

    /**
     * Standard event type: READ
     */
    public static String READ = "Read";
    /**
     * Standard event type: READ_FINISHED
     */
    public static String READ_FINISHED = "ReadFinished";
    /**
     * Standard event type: COPY_FROM
     */
    public static String COPY_FROM = "CopyFrom";
    /**
     * Standard event type: PASTE_TO
     */
    public static String PASTE_TO = "PasteTo";
    /**
     * Standard event type: FILE_RENAMED_FROM
     */
    public static String FILE_RENAMED_FROM = "FileRenamedFrom";
    /**
     * Standard event type: FILE_RENAMED_TO
     */
    public static String FILE_RENAMED_TO = "FileRenamedTo";
    /**
     * Standard event type: FILE_MODIFIED
     */
    public static String FILE_MODIFIED = "FileModified";
    public static String TIMED = "Timed";
    public static String REMOTE;
    public static String ECHO;
    public static String EVENT_TYPE_TAIL_CHANGE = "TailChange";

    /**
     * Timestamp of the eventType.
     */
    private Date date;
    /**
     * Local localDbId of the object associated with the eventType.
     */
    private String localObjectId;
    /**
     * The content of the eventType.
     */
    private Object content;
    /**
     * The name of the event.
     */
    private String eventType;
    /**
     * The localDbId of the eventType in the local database.
     * ({@link hu.sztaki.dsd.visionair.a2a3d.localStorage.DataStore})
     */
    private int localDbId;
    /**
     * The eventType is triggered by the
     * {@link hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSource} with this
     * localDbId (an implementation of the triggering can be found in
     * {@link hu.sztaki.dsd.visionair.a2a3d.eventSources.SimpleEventSource}).
     */
    //public String triggeredBy;
    /**
     * Additional parameters set by the triggering
     * {@link hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSource}.
     */
    //public String triggerParams;
    /**
     * Identifies the source software environment of the eventType.
     */
    private String sourceMachineId;

    public Event() {
    }

    /**
     * Creates a new instance of Event object.
     *
     * @param localObjectId Local localDbId of the associated object (normally
     * the path identifying a given document/url)
     * @param eventType the name of the eventType
     * @param content additional parameters (e.g. the content of the copied
     * text)
     */
    public Event(String localObjectId, String eventType, Object content) {
        this.date = new Date();
        this.localObjectId = localObjectId;
        this.eventType = eventType;
        this.content = content;
        this.sourceMachineId = A2A3D.machineId;
    }

    /**
     * Creates a new instance of Event object. This is used whenever the
     * triggering {@link hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSource}
     * is important.
     *
     * @param localObjectId Local localDbId of the associated object (normally
     * the path identifying a given document/url)
     * @param eventType The name of the eventType.
     * @param content Additional parameters (e.g. the content of the copied
     * text).
     * @param triggeredBy The localDbId of the
     * {@link hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSource} triggering
     * this eventType.
     * @param triggerParams Additional parameters of the
     * {@link hu.sztaki.dsd.visionair.a2a3d.eventSources.EventSource} triggering
     * this eventType.
     */
    /*
     * public Event(String localObjectId, String event, String content, String
     * triggeredBy, String triggerParams) { this.timestamp = new
     * Date().getTime(); this.localObjectId = localObjectId; this.eventType =
     * event; this.content = content; this.triggeredBy = triggeredBy;
     * this.triggerParams = triggerParams; }
     */
    /**
     * Called by the DataStore. It will initiate the Event objects based on
     * values stored in the database.
     *
     * @param localDbId Id of the Event in the local database
     * @param csId The associated CSpace localDbId to the eventType.
     * @param objId The localDbId of the object that the eventType belongs to.
     * @param eventType The name of the eventType.
     * @param name Not used.
     * @param date Timestamp of the eventType.
     * @param message Not used.
     */
    public Event(int localDbId, String event, String name, Date date, String message) {
        this.localDbId = localDbId;
        this.eventType = event;
        this.date = date;
    }

    /**
     * Setter method of the field localDbId.
     *
     * @param localDbId The new value of localDbId.
     */
    public void setLocalDbId(int localDbId) {
        this.localDbId = localDbId;
    }

    /**
     * Serialises the Event for debugging purposes.
     *
     * @return The human readable form of the Event.
     */
    @Override
    public String toString() {
        String retVal = "Event:\n";
        retVal += "\tdate: " + date;
        retVal += "\n\tlocalId: " + localObjectId;
        retVal += "\n\tcontent: " + content;
        retVal += "\n\tevent: " + eventType;
        retVal += "\n\tlocalDbId: " + localDbId;
        //retVal += "\n\ttriggerParams: " + triggerParams;
        retVal += "\n\tmachineId: " + sourceMachineId;
        return retVal;
    }

    @XmlElement(name = "content")
    public String getContent() throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        StringWriter sw = new StringWriter();
        mapper.writeValue(sw, content);
        sw.flush();
        return sw.toString();
    }

    @Id
    @GeneratedValue(generator = "increment")
    @GenericGenerator(name = "increment", strategy = "increment")
    public int getLocalDbId() {
        return localDbId;
    }

    @Temporal(TemporalType.TIMESTAMP)
    @Column(name = "eventdate")
    public Date getDate() {
        return date;
    }

    public void setDate(Date date) {
        this.date = date;
    }

    public void setContent(String content) {
        this.content = content;
    }

    public void setEventType(String eventType) {
        this.eventType = eventType;
    }

    public void setLocalObjectId(String localObjectId) {
        this.localObjectId = localObjectId;
    }

    public void setSourceMachineId(String sourceMachineId) {
        this.sourceMachineId = sourceMachineId;
    }

    public String getEventType() {
        return eventType;
    }

    public String getLocalObjectId() {
        return localObjectId;
    }

    public String getSourceMachineId() {
        return sourceMachineId;
    }

}
