/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package hu.sztaki.dsd.visionair.a2a3d.eventSources;

import hu.sztaki.dsd.visionair.a2a3d.Event;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.eventFilters.EventFilter;
import java.util.ArrayList;
import java.util.HashMap;
import javax.xml.bind.annotation.XmlElement;

/**
 * <p>EventSource interface. The whole system consists of a network of 
 * <code>EventSource</code>-s. They are responsible for creating the events. 
 * They have a trigger queue and in the case of newly generated events they has 
 * to trigger other <code>EventSource</code>-s subscribed for them. Even the 
 * communication between the A2A3D agents are implemented via 
 * <code>EventSource</code>-s posting events externally. 
 * 
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt; (MTA
 * SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
public interface EventSource {

    /**
     * The initial status of an <code>EventSource</code>.
     */
    static String STATUS_NONE = "None";

    /**
     * The status of the <code>EventSource</code> is initiating and cannot send
     * events yet.
     */
    static String STATUS_INITIATING = "Initiating";

    /**
     * The status of the <code>EventSource</code> is running and can send and
     * accept events.
     */
    static String STATUS_RUNNING = "Running";

    /**
     * The status of the <code>EventSource</code> is stopping, cannot accept 
     * new events, the trigger queue is being shut down.
     */
    static String STATUS_STOPPING = "Stopping";
    
    /**
     * The status of the <code>EventSource</code> is error, cannot accept 
     * new events, the trigger queue is not working.
     */
    static String STATUS_ERROR = "Error";

    /**
     * Called at shutdown to tell the EventSource to cleanup after itself if it
     * is necessary.
     */
    public void destroy();
    
    /**
     *
     */
    public void shutDownNotificationQueues();
    
    /**
     *
     */
    public void waitForShutDownNotificationQueues();
    
    /**
     * Waits while the EventSource kills it's notificators and safely cleans up.
     */
    public void waitForDestroy();

    /**
     * Adds a new <code>EventSource</code> instance to the trigger queue of the 
     * actual one.
     * 
     * @param id identifier of the <code>EventSource</code> that will be triggered.
     * @param es the <code>EventSource</code> instance that will be triggered.
     */
    public void addToQueue(String id, EventSource es);

    /**
     * Called at init time to initialise the EventSource.
     */
    public void init();

    /**
     *
     * @return the ClassName of the EventSource.
     */
    @XmlElement(name = "className")
    public String getClassName();

    /**
     * Returns the actual status of the <code>EventSource</code>.
     * <p>
     * It can be {@link STATUS_NONE}, {@link STATUS_INITIATING}, {@link STATUS_RUNNING} or
     * {@link STATUS_STOPPING}.
     * 
     * @return the actual status of the <code>EventSource</code>.
     */
    @XmlElement(name = "status")
    public String getStatus();

    /**
     * Returns the notification queue of the <code>EventSource</code> instance.
     * <p>
     * Each <code>EventSource</code> in the notification queue has a dedicated
     * {@link ThreadNotificator} instance. These are identified by strings
     * coming from their configurations. The notification queue is implemented
     * as a <code>HashMap</code> where the key is this identifier and the stored
     * values are the <code>ThreadNotificator</code> instances.
     * 
     * @return the notification queue of the <code>EventSource</code> instance.
     */
    @XmlElement(name = "triggerQueue")
    public HashMap<String, ThreadNotificator> getTriggerQueue();
    
    /**
     * Called by other <code>EventSource</code> instances where this instance is
     * in their notification queue.
     * @param event the incoming event. The event sent by other 
     * <code>EventSource</code> instances.
     */
    public boolean onHookEvent(Event event);
    
    @XmlElement(name = "filterList")
    public ArrayList<EventFilter> getEventFilterList();
}
