/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package hu.sztaki.dsd.visionair.a2a3d.eventSources;

import hu.sztaki.dsd.visionair.a2a3d.Event;
import java.util.Date;
import org.apache.commons.configuration.SubnodeConfiguration;
import static org.quartz.JobBuilder.newJob;
import org.quartz.JobDetail;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import static org.quartz.SimpleScheduleBuilder.simpleSchedule;
import org.quartz.SimpleTrigger;
import static org.quartz.TriggerBuilder.newTrigger;
import org.quartz.impl.StdSchedulerFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>
 * Triggers in given time intervals. The triggering timeout must be indicated
 * among the parameters in seconds (timeout). More than one SimpleTimedEvent
 * event trigger might be instantiated at the same time with different
 * frequencies.
 *
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt; (MTA
 * SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
class QuartzSimpleTimedEventSource extends SimpleEventSource {

    /**
     * Logger of the class.
     */
    private static Logger logger = LoggerFactory.getLogger(QuartzSimpleTimedEventSource.class);
    /**
     * Scheduler of the Quartz jobs.
     */
    private Scheduler scheduler = null;
    /**
     * Timeout of job execution in sec.
     */
    private int timeout;

    /**
     * Instantiates a new QuartzSimpleTimedEventSource.
     *
     * @param xmlConfig Configuration parameters.
     */
    public QuartzSimpleTimedEventSource(SubnodeConfiguration xmlConfig) {
        super(xmlConfig);
    }

    /**
     * Initiates the QuartzSimpleTimedEventSource. It start to trigger its queue
     * with the set timeout.
     */
    public void init() {
        setStatus(STATUS_INITIATING);
        try {
            // Grab the Scheduler instance from the Factory 
            scheduler = StdSchedulerFactory.getDefaultScheduler();

            // and start it off
            scheduler.start();
            
            SubnodeConfiguration xmlConfig = getConfiguration();
            
            timeout = xmlConfig.getInt("Parameters.Timeout");
            logger.debug("The trigger timeout for the SimpleTimedEventSource(" + xmlConfig.getString("Id") + ") is: " + timeout);
            
            SimpleTrigger trigger = (SimpleTrigger) newTrigger().withIdentity(xmlConfig.getString("Id"), "group1").startAt(new Date()).withSchedule(simpleSchedule().withIntervalInSeconds(timeout).repeatForever()).build();
            
            JobDetail job = newJob(QuartzJob.class).withIdentity(xmlConfig.getString("Id"), "group1").build();
            
            job.getJobDataMap().put("Id", xmlConfig.getString("Id"));
            
            scheduler.scheduleJob(job, trigger);
        } catch (SchedulerException ex) {
            logger.error(ex.toString());
        }
        super.init();
    }

    /**
     * Called at shutdown time.
     */
    public void destroy() {
        setStatus(STATUS_STOPPING);
        try {
            scheduler.shutdown();
        } catch (SchedulerException ex) {
            logger.error(ex.toString());
        }
        super.destroy();
    }

    /**
     * This function is called by the QuartzJob instances.
     *
     * @param id The id of the QuartzJob instance.
     */
    public static void trigger(String id) {
        try {
            SimpleTimedEvent se = new SimpleTimedEvent();
            logger.debug("TRIGGER ID: " + id);
            QuartzSimpleTimedEventSource qstes = (QuartzSimpleTimedEventSource) se.getInstance(id);
            logger.debug("QuartzSimpleTimedEventSource: " + qstes.getStatus());
            qstes.notifyQueue(new Event("", Event.TIMED, ""));
        } catch (NoEventSourceAviableError ex) {
            logger.error(ex.toString());
        }
    }
}
