/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package hu.sztaki.dsd.visionair.a2a3d.eventSources;

import hu.sztaki.dsd.visionair.a2a3d.Event;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.eventFilters.EventFilter;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.eventFilters.EventFilterFactory;
import hu.sztaki.dsd.visionair.a2a3d.eventSources.eventFilters.NoEventFilterAvailableError;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import javax.xml.bind.annotation.XmlElement;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A simple generic EventSource, which implements the functions necessary to
 * handle the listener queues of an event trigger. Our EventSource
 * implementations extend this class.
 *
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt;
 * (MTA SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
public class SimpleEventSource implements EventSource {

    /**
     * Logger of the class.
     */
    private static Logger logger = LoggerFactory.getLogger(SimpleEventSource.class);

    /**
     * The status of the EventSource. Possible values are: STATUS_NONE,
     * STATUS_RUNNING, STATUS_STOPPING. Initially it is set to STATUS_NONE.
     */
    private String status = STATUS_NONE;
    /**
     * Configuration parameters set at instantiation time.
     */
    private SubnodeConfiguration xmlConfig = null;
    /**
     * The queue of the EventSources waiting for notifications.
     */
    private HashMap<String, ThreadNotificator> triggerQueue = new HashMap<String, ThreadNotificator>();
    /**
     * Event filters for the incoming event queue.
     */
    private ArrayList<EventFilter> eventFilterList = new ArrayList<EventFilter>();

    public SimpleEventSource() {
    }

    public SimpleEventSource(SubnodeConfiguration xmlConfig) {
        this.xmlConfig = xmlConfig;
    }

    /**
     * Init function.
     */
    @Override
    public void init() {
        status = STATUS_INITIATING;
        if (xmlConfig != null) {
            List eventFilterClassNameList = xmlConfig.getList("EventFilters.EventFilter.ClassName");
            logger.debug("Number of found EventFilters: " + eventFilterClassNameList.size());

            for (int listCount = 0; listCount < eventFilterClassNameList.size(); listCount++) {
                try {
                    Class eventFilterClass = Class.forName("" + eventFilterClassNameList.get(listCount));
                    if (EventFilterFactory.class.isAssignableFrom(eventFilterClass)) {

                        // add the filter to the cache
                        try {
                            logger.info("Initializing the EventFilter: " + eventFilterClass.getName());
                            EventFilterFactory eff = (EventFilterFactory) eventFilterClass.newInstance();
                            EventFilter ef;
                            try {
                                ef = eff.getInstance(xmlConfig.configurationAt("EventFilters.EventFilter(" + listCount + ")"));
                                ef.init();
                                this.eventFilterList.add(ef);
                                logger.info("Initializing of EventFilter: " + listCount + " - Finished");
                            } catch (NoEventFilterAvailableError ex) {
                                logger.error(ex.toString());
                            }

                        } catch (InstantiationException ex) {
                            logger.error(ex.toString());
                        } catch (IllegalAccessException ex) {
                            logger.error(ex.toString());
                        }

                    }
                } catch (ClassNotFoundException ex) {
                    logger.error(ex.toString());
                    logger.error("Invalid EventFilter className");
                    System.exit(1);
                }
            }
        }
        status = STATUS_RUNNING;
    }

    /**
     * Destroy function.
     */
    @Override
    public void destroy() {
        status = STATUS_STOPPING;
        Iterator iterator = triggerQueue.keySet().iterator();
        while (iterator.hasNext()) {
            ThreadNotificator tn = triggerQueue.get(iterator.next());
            tn.notificatorShutDown();
        }
    }

    /**
     * Adds a new DataCollector to the triggerQueue of the EventSource
     *
     * @param id Id of the EventSource waiting for trigger messages
     * @param es The EventSource waiting for trigger messages
     */
    @Override
    public void addToQueue(String id, EventSource es) {
        if (!triggerQueue.containsKey(id)) {
            ThreadNotificator tn = new ThreadNotificator(es);
            triggerQueue.put(id, tn);
            Thread notificationThread = new Thread(tn);
            notificationThread.start();
            while (tn.getWorkingThread() == null) {
                try {
                    Thread.sleep(10);
                } catch (InterruptedException ex) {
                }
            }
            //logger.debug("Notification thread started (" + id + ")");
        }
    }

    /**
     * Iterates through the list of the ThreadNotificators and sends them a
     * TriggerMessage.
     *
     * @param event The TriggerMessage to be sent to the EventSources.
     */
    public void notifyQueue(Event event) {
        //logger.debug("notify queue running");
        Iterator queueIterator = triggerQueue.keySet().iterator();
        while (queueIterator.hasNext()) {
            //logger.debug("triggerQueue addevent running");
            triggerQueue.get(queueIterator.next()).addEvent(event);
        }
    }

    @Override
    public String getClassName() {
        return this.getClass().getName();
    }

    @Override
    public String getStatus() {
        return status;
    }

    @Override
    public HashMap<String, ThreadNotificator> getTriggerQueue() {
        return triggerQueue;
    }

    @Override
    public boolean onHookEvent(Event event) {
        int valueCheck = EventFilter.NONE;
        for (EventFilter ef : eventFilterList) {
            valueCheck = ef.check(event);
            if (valueCheck != EventFilter.NONE) {
                break;
            }
        }
        // If the value is NONE or ACCEPT, accept the event
        if (valueCheck != EventFilter.DENY) {
            return true;
        } else {
            return false;
        }
    }

    public void setStatus(String status) {
        this.status = status;
    }

    @Override
    public void waitForDestroy() {
        waitForShutDownNotificationQueues();
    }

    @Override
    public void shutDownNotificationQueues() {
        Iterator queueIterator = triggerQueue.keySet().iterator();
        while (queueIterator.hasNext()) {
            ThreadNotificator tn = (ThreadNotificator) triggerQueue.get(queueIterator.next());
            tn.notificatorShutDown();
        }
    }

    @Override
    public void waitForShutDownNotificationQueues() {
        Iterator queueIterator = triggerQueue.keySet().iterator();
        while (queueIterator.hasNext()) {
            ThreadNotificator tn = (ThreadNotificator) triggerQueue.get(queueIterator.next());
            while (tn.getWorkingThread().isAlive()) {
                try {
                    tn.getWorkingThread().join();
                } catch (InterruptedException ex) {
                }
            }
        }
    }

    public SubnodeConfiguration getConfiguration() {
        return xmlConfig;
    }

    public void setConfiguration(SubnodeConfiguration xmlConfig) {
        this.xmlConfig = xmlConfig;
        return;
    }

    @XmlElement(name = "filterList")
    @Override
    public ArrayList<EventFilter> getEventFilterList() {
        return eventFilterList;
    }
}
