/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package hu.sztaki.dsd.visionair.a2a3d.eventSources;

import hu.sztaki.dsd.visionair.a2a3d.Event;
import static java.lang.Thread.interrupted;
import java.util.LinkedList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The notifications of the EventSources happen in separate Threads. These
 * threads run the ThreadNotificators which handle the actual notification.
 *
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt; (MTA
 * SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
public class ThreadNotificator implements Runnable {

    /**
     * The EventSource to be notified.
     */
    private final EventSource es;
    /**
     * The list of events to be sent to the EventSources.
     */
    private final LinkedList<Event> eventList = new LinkedList<Event>();
    /**
     * Object for synchronisation purposes.
     */
    private final Object itemArrivedLock = new Object();
    /**
     * boolean value indicating if the notificator is requested to shut down.
     */
    private boolean notificatorShutDown = false;
    /**
     * The thread processing the incoming events.
     */
    private Thread workingThread = null;

    /**
     * Logger of the class.
     */
    private static Logger logger = LoggerFactory.getLogger(ThreadNotificator.class);

    /**
     * Instantiates a new ThreadNotificator.
     *
     * @param es EventSource to be notified.
     * @param event TriggerMessage to be sent to the DataCollector.
     */
    public ThreadNotificator(EventSource es) {
        this.es = es;
    }

    /**
     * Adds a notification event to the event queue.
     *
     * @param event
     */
    public void addEvent(Event event) {
        if (!notificatorShutDown && event != null) {
            synchronized (eventList) {
                eventList.add(event);
                //logger.debug("eventList.add(event)");
            }
            //logger.debug("Event added (" + eventList.size() + ")");
            synchronized (itemArrivedLock) {
                itemArrivedLock.notifyAll();
                //logger.debug("itemArrivedLock.notifyAll()");
            }
            //logger.debug("Thread notified.");
        }
    }

    /**
     * Consumes an event from the notification queue
     *
     * @return Event to be served from the top of the notification event queue
     */
    private Event pop() {
        synchronized (eventList) {
            if (!eventList.isEmpty()) {
                return eventList.remove(0);
            }
        }
        return null;
    }

    /**
     * Triggers the EventSource with the TriggerMessage.
     */
    @Override
    public void run() {
        Event event;
        workingThread = Thread.currentThread();
        //logger.debug("Run called.");
        while (!notificatorShutDown) {
            //logger.debug("Consuming events.");
            while ((event = pop()) != null) {
                //logger.debug("Event found.");
                es.onHookEvent(event);
                //logger.debug("Sources notified.");
            }
            try {
                synchronized (itemArrivedLock) {
                    //logger.debug("Waiting");
                    if (!notificatorShutDown) {
                        itemArrivedLock.wait();
                    }
                    //logger.debug("Wake up");
                }
            } catch (InterruptedException ex) {
                interrupted();
            }
        }
        //logger.debug("Shut down");
    }

    /**
     * Returns the working thread processing the incoming events. It might be
     * necessary to synchronise at shutdown time.
     *
     * @return
     */
    public Thread getWorkingThread() {
        return this.workingThread;
    }

    /**
     * Shuts down the notificator. It will consume all events from the queue
     * before shutting down!
     */
    public void notificatorShutDown() {
        logger.debug("shutdown called");
        notificatorShutDown = true;
        synchronized (itemArrivedLock) {
            itemArrivedLock.notifyAll();
        }
        //logger.debug("shutdown notification sent");
    }

    /**
     * Returns the notification queue.
     * @return LinkedList of Events to be sent to the triggered EventSource.
     */
    public LinkedList<Event> getEventList() {
        return eventList;
    }
}
