/*
 * Copyright 2014 MTA SZTAKI - DSD. All rights reserved.
 * MTA SZTAKI - DSD PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package hu.sztaki.dsd.visionair.a2a3d.localStorage;

import hu.sztaki.dsd.visionair.a2a3d.Event;
import java.sql.SQLException;
import java.util.ArrayList;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.cfg.Configuration;
import org.hibernate.service.ServiceRegistry;
import org.hibernate.service.ServiceRegistryBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>
 * Persistent datastore for the A2A3D installation. Now it is used as a
 * temporary datastore of events collected by the DataCollectors. The collected
 * eventType will be stored in this local store (it is considered to be faster
 * and more reliable than the link between the server and the sensor) and then
 * forwarded to the sensor until it successfully processes it. As database a
 * simple SQLite database is used.
 *
 * The configuration of the datastore is in the hiberante.cfg.xml.
 *
 * @author Zoltan Toth &lt;<a
 * href="mailto:toth.zoltan@sztaki.mta.hu">toth.zoltan@sztaki.mta.hu</a>&gt;
 * (MTA SZTAKI, <a href="http://dsd.sztaki.hu">DSD</a>)
 */
public class DataStore {

    /**
     * Logger of the class.
     */
    private static Logger logger = LoggerFactory.getLogger(DataStore.class);
    private SessionFactory sessionFactory;

    /**
     * Connects to the db, initialises if it is not exists yet.
     *
     * @param xmlconfig Configuration of the database (see readme.txt)
     */
    public DataStore() {
        // A SessionFactory is set up once for an application
        Configuration configuration = new Configuration();
        configuration.configure();
        ServiceRegistry serviceRegistry = new ServiceRegistryBuilder().applySettings(configuration.getProperties()).buildServiceRegistry();
        sessionFactory = configuration.buildSessionFactory(serviceRegistry);
    }

    /**
     * Releases the connection to the database. Called by the ShutDown hook of
     * the agent.
     */
    public void close() {
        if (sessionFactory != null) {
            sessionFactory.close();
        }
    }

    /**
     * After successfully sending an eventType the stored version of it will be
     * deleted from the database calling this method.
     *
     * @param event The eventType to be deleted from the database.
     * @throws java.sql.SQLException
     */
    public synchronized void deleteEvent(Event event) throws SQLException {
        logger.debug("Deleting from local db store. Id: " + event.getLocalDbId());
        Session session = sessionFactory.openSession();
        session.beginTransaction();
        session.delete(event);
        session.getTransaction().commit();
        session.close();

        logger.debug("Deleted from local db store. Id: " + event.getLocalDbId());
    }

    /**
     * Requests the ArrayList of unsent events from the database.
     *
     * @return ArrayList of Event objects containing all events in the database
     * @throws java.sql.SQLException
     */
    public synchronized ArrayList<Event> getEvents() throws SQLException {
        // now lets pull events from the database and list them
        Session session = sessionFactory.openSession();
        session.beginTransaction();
        @SuppressWarnings("unchecked")
        ArrayList<Event> result = (ArrayList<Event>) session.createQuery("from Event order by eventdate").list();
        session.getTransaction().commit();
        session.close();
        return result;
    }

    /**
     * Stores a new eventType into the database.
     *
     * @param collectedEvent the Event to be stored in the database
     */
    public synchronized void storeEvent(Event collectedEvent) {
        Session session = sessionFactory.openSession();
        session.beginTransaction();
        session.save(collectedEvent);
        session.getTransaction().commit();
        session.close();
    }

    public boolean isClosed() {
        if (sessionFactory != null) {
            return sessionFactory.isClosed();
        }
        return true;
    }

}
